;;;;;; FASL library for Scheme48                         -*- Scheme -*-
;;;;;; FASL file dumping & loading

;;; Taylor Campbell wrote this code; he places it in the public domain.

(define fasl-dump-level 4)

(define $fasl-structures (make-fluid #f))

(define (dump-fasl filename defs structs)
  (let ((structs (extract-fasl-structures defs structs)))
    (load-extra-packages structs)
    (let ((filename (namestring filename #f 'fasl)))
      (call-with-file-dumpster filename fasl-dump-encoder
        (lambda (dumpster)
          (dump-integer dumpster fasl-dump-level)
          (let-fluid $fasl-structures structs
            (lambda ()
              (for-each (lambda (package)
                          (dump dumpster package)
                          (fasdump-package dumpster package))
                        (collect-packages structs
                          (lambda (package)
                            (any-structure-has-package? structs
                                                        package))))))
          (dump dumpster #f)
          (for-each (lambda (name.value)
                      (dump dumpster (car name.value))
                      (dump dumpster (cdr name.value)))
                    defs)
          (dump dumpster #f))))))

(define (extract-fasl-structures defs structs)
  (let loop ((defs defs)
             (structs structs))
    (if (null? defs)
        structs
        (loop (cdr defs)
              (let ((value (cdar defs)))
                (if (structure? value)
                    (cons value structs)
                    structs))))))

(define (load-extra-packages structs)
  (for-each compile-and-load-package
            (collect-packages
             ;; If we were to pass simply STRUCTS here, we'd get the
             ;; empty list, because it wouldn't bother descending at
             ;; all.  Instead we need to start it off with what all of
             ;; the structures open.
             (apply append
                    (map (lambda (struct)
                           (package-opens (structure-package struct)))
                         structs))
             (lambda (package)
               (not (or (package-loaded? package)
                        (any-structure-has-package? structs
                                                    package)))))))

(define (compile-and-load-package package)
  (load-package package compile-package))

(define (load-package package package->template)
  (with-interaction-environment package
    (lambda ()
      (invoke-closure
       (make-closure (package->template package)
                     (package-uid package)))))
  (set-package-loaded?! package #t)
  (walk-population check-structure
                   (package-clients package)))

(define (any-structure-has-package? structs package)
  (let loop ((structs structs))
    (and (pair? structs)
         (or (eq? (structure-package (car structs))
                  package)
             (loop (cdr structs))))))

(define (fasl-dump-encoder obj)
  (cond ((structure? obj) (values #f 'structure structure-dump-encoder))
        ((interface? obj) (values #f 'interface interface-dump-encoder))
        ((package?   obj) (values #f 'package     package-dump-encoder))
        (else             (usual-dump-encoder obj))))

(define structure-dump-encoder
  (lambda (dumpster struct)
    (dump dumpster (structure-interface struct))
    (dump dumpster (structure-package struct))))

(define interface-dump-encoder
  (lambda (dumpster interface)
    (for-each-declaration (lambda (external-name internal-name type)
                            (dump dumpster external-name)
                            (dump dumpster internal-name)
                            (dump dumpster
                                  (if (eq? type undeclared-type)
                                      #f
                                      (type->sexp type #t))))
                          interface)
    (dump dumpster #f)))

(define package-dump-encoder
  (lambda (dumpster package)
    (dump dumpster (package-name package))
    (dump dumpster (map structure->dump-token (package-opens package)))
    (dump dumpster
          (map (lambda (access)
                 (cons (car access)
                       (structure->dump-token (cdr access))))
               (package-accesses package)))
    (dump dumpster (package-unstable? package))
    (dump dumpster (package-file-name package))
    (dump dumpster (package-clauses package))))

(define (fasdump-package dumpster package)
  (compile-and-dump-package package
    (lambda (dump-encoder receiver)
      (let ((sub-dumpster (make-dumpster dump-encoder)))
        (receiver sub-dumpster)
        (dump dumpster
              (call-with-string-output-port
                (lambda (port)
                  (write-dump sub-dumpster port))))))))

;;; If it's in this FASL, we just directly refer to it; otherwise we'll
;;; refer to it by name.

(define (structure->dump-token struct)
  (if (memq struct (fluid $fasl-structures))
      struct
      (or (structure-name struct)
          (structure-modifier struct)
          (error "unable to dump reference to anonymous structure"
                 struct))))

;++ This is a heuristic crock.

(define (structure-modifier struct)
  (and-let* ((package (structure-package struct))
             (name (package-name package))
             (binding (package-lookup (config-package) name))
             ((not (eq? (binding-type binding) syntax-type)))
             (loc (binding-place binding))
             ((location? loc))
             ((location-defined? loc))
             (obj (contents loc))
             ((not (eq? obj struct)))   ; Probably unnecessary.
             ((structure? obj))
             ((eq? (structure-package obj) package)))
    (receive (exported renamed)
             (decompose-modified-interface
              (structure-interface struct))
      `(MODIFIED ,name
                 ,(make-simple-interface #f exported)
                 ,@renamed))))

(define (decompose-modified-interface interface)
  (let ((exported '())
        (renamed '()))
    (for-each-declaration (lambda (name base-name type)
                            (set! exported
                                  (cons (if (eq? type undeclared-type)
                                            base-name
                                            (list base-name type))
                                        exported))
                            (if (not (eq? name base-name))
                                (set! renamed
                                      (cons (list base-name name)
                                            renamed))))
                          interface)
    (values exported renamed)))



(define (load-fasl filename)
  (call-with-file-retriever (namestring filename #f 'fasl)
      fasl-dump-decoder
    (lambda (retriever)
      (if (not (= (retrieve-integer retriever)
                  fasl-dump-level))
          (error "FASL format mismatch"
                 filename
                 '(try recompiling)))
      (let* ((packages (retrieve-fasl-packages retriever))
             (structs (retrieve-fasl-definitions retriever)))
        (load-fasl-structures structs packages)))))

(define (retrieve-fasl-packages retriever)
  (let loop ((packages '()))
    (cond ((retrieve retriever)
           => (lambda (package)
                (loop (cons (cons package (retrieve retriever))
                            packages))))
          (else packages))))

(define (retrieve-fasl-definitions retriever)
  (let loop ((structs '()))
    (cond ((retrieve retriever)
           => (lambda (name)
                (let ((value (retrieve retriever)))
                  (environment-define! (config-package)
                                       name value)
                  (cond ((structure? value)
                         (note-structure-name! value name)
                         (loop (cons value structs)))
                        ((interface? value)
                         (note-interface-name! value name)
                         (loop structs))
                        (else
                         (loop structs))))))
          (else structs))))

(define (load-fasl-structures structs packages)
  (for-each (lambda (package)
              (cond ((assq package packages)
                     => (lambda (probe)
                          (load-package package
                                        (lambda (p)
                                          (fasload-package (cdr probe)
                                                           p)))))
                    (else
                     (compile-and-load-package package))))
            (collect-packages structs
                              (lambda (p)
                                (not (package-loaded? p))))))

(define (fasl-dump-decoder key)
  (cond ((eq? key 'structure) structure-dump-decoder)
        ((eq? key 'interface) interface-dump-decoder)
        ((eq? key 'package)     package-dump-decoder)
        (else (usual-dump-decoder key))))

(define structure-dump-decoder
  (lambda (retriever port register!)
    (let* ((interface (retrieve retriever))
           (package (retrieve retriever))
           (struct (make-structure package interface)))
      (register! struct)
      struct)))

(define interface-dump-decoder
  (lambda (retriever port register!)
    (let loop ((exports '())
               (renames '()))
      (cond ((retrieve retriever)
             => (lambda (external-name)
                  (let* ((internal-name (retrieve retriever))
                         (type
                          (cond ((retrieve retriever)
                                 => (lambda (sexp)
                                      (sexp->type sexp #t)))
                                (else undeclared-type))))
                    (loop (cons (list internal-name type)
                                exports)
                          (if (eq? internal-name external-name)
                              renames
                              (cons (list internal-name
                                          external-name)
                                    renames))))))
            (else
             (let* ((simple (make-simple-interface #f exports))
                    (interface (if (null? renames)
                                   simple
                                   ((make-modified-interface-maker
                                     `((RENAME ,@renames)))
                                    simple))))
               (register! interface)
               interface))))))

(define package-dump-decoder
  (lambda (retriever port register!)
    (let* ((name (retrieve retriever))
           (opens (retrieve retriever))
           (accesses (retrieve retriever))
           (unstable? (retrieve retriever))
           (filename (retrieve retriever))
           (clauses (retrieve retriever))
           (package (make-package (lambda () opens)
                                  (lambda () accesses)
                                  unstable?
                                  (make-reflective-tower clauses name)
                                  filename
                                  clauses
                                  #f    ; uid
                                  name)))
      (register! package)
      (set! opens (map dump-token->structure opens))
      (set! accesses (map (lambda (access)
                            (cons (car access)
                                  (dump-token->structure
                                   (cdr access))))
                          accesses))
      package)))

(define (fasload-package dump-string package)
  (retrieve-package-template package
    (lambda (dump-decoder receiver)
      (receiver (make-retriever (make-string-input-port dump-string)
                                dump-decoder)))))

(define (dump-token->structure token)
  (cond ((structure? token)
         token)
        ((symbol? token)
         (get-structure token))
        ((and (pair? token)
              (eq? (car token) 'modified))
         (destructure (( (name interface . renamed) (cdr token)))
           (let ((struct (make-structure (structure-package
                                          (get-structure name))
                                         interface)))
             (if (null? renamed)
                 struct
                 (make-modified-structure struct
                                          `((RENAME ,@renamed)))))))
        (else
         (error "invalid external structure reference" token))))

(define (make-reflective-tower clauses name)
  ((environment-ref (config-package)
                    (string->symbol ".make-reflective-tower."))
   clauses
   name))
